#
# ZopeTree
#
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public License
# (MPL)
#

from AccessControl import allow_module

__allow_access_to_unprotected_subobjects__ = 1
__roles__ = None

# make ZopeTree module accessible from PythonScript and ZPT
from Products.ZopeTree.ZopeTree import ZopeTree, Node  # noqa
from Products.ZopeTree.TreeObjectWrapper import TreeObjectWrapper  # noqa
allow_module('Products.ZopeTree')



#
# ZopeTree
#
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public License
# (MPL)
#

"""
$Id: __init__.py,v 1.5 2003/03/15 17:40:06 philipp Exp $
"""

# make sure, 'from Interface import Interface' will work...
try:
    from Interface import Interface
except ImportError:
    # Zope <2.6
    import Interface
    Interface.Interface = Interface.Base

# make sure 'from AccessControl import allow_module' works...
try:
    from AccessControl import allow_module
except ImportError:
    # Zope <2.5
    import AccessControl
    from AccessControl import ModuleSecurityInfo
    def allow_module(module_name):
        ModuleSecurityInfo(module_name).setDefaultAccess(1)
        dot = module_name.find('.')
        while dot > 0:
            ModuleSecurityInfo(module_name[:dot]).setDefaultAccess(1)
            dot = module_name.find('.', dot + 1)

__allow_access_to_unprotected_subobjects__ = 1
__roles__ = None

# make ZopeTree module accessible from PythonScript and ZPT
from ZopeTree import ZopeTree, Node
from TreeObjectWrapper import TreeObjectWrapper
allow_module('Products.ZopeTree')



#
# TreeObjectWrapper
#
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public
# License (MPL)
#

"""
$Id: TreeObjectWrapper.py,v 1.5 2003/04/30 10:01:55 philipp Exp $
"""

import posixpath


class TreeObjectWrapper:
    """
    A wrapper around Zope objects to be more flexible when displaying
    them with ZopeTree.

    Extend this class with methods filtering the return list of
    getChildren() to your own needs.
    """

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, obj, path=''):
        self.path = path
        self.object = obj
        self.id = obj.getId()
        self.folderish = obj.isPrincipiaFolderish

    def getChildren(self):
        if not self.folderish:
            return []
        children = []
        objectIds = self.object.objectIds()
        objectIds.sort()
        for id in objectIds:
            path = posixpath.join(self.path, id)
            obj = getattr(self.object, id)
            children.append(TreeObjectWrapper(obj, path))
        return children

    def getFolderishChildren(self):
        return [child for child in self.getChildren() if child.folderish]



#
# IZopeTree
#
# Interface for Zope trees
#
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public
# License (MPL)
#

from zope.interface import Interface, Attribute


class INode(Interface):

    object = Attribute("""
        The object that is being wrapped.
        """)

    depth = Attribute("""
        The positional depth of this node in the tree.
        """)

    expanded = Attribute("""
        True if this node is expanded.
        """)

    def __init__(object, depth, id_attr, children_attr, expanded_nodes=[]):
        """
        'object' is the object that is being wrapped

        'depth' is the positional depth of the node in the tree

        'id_attr' should point to an attribute/method name of 'object'
        (and all other wrapped objects), giving each object a unique
        id in the tree.

        'children_attr' should point to an attribute/method name of
        'object' containing/resulting object's children.

        'expanded_nodes' is a list containing ids of nodes that are
        expanded from the beginning.
        """

    def expand(recursive=0):
        """
        Expand this node.

        'recursive' can be set to true to expand all children nodes as
        well
        """

    def collapse():
        """
        Collapse this node.
        """

    def getId():
        """
        Return the object's id in the tree.
        """

    def hasChildren():
        """
        Return true if we have children.
        """

    def getChildrenNodes():
        """
        Return a sequence of children nodes if the node is expanded.
        """

    def getFlatNodes():
        """
        Return a flat list of nodes in the tree. Children of expanded
        nodes are shown.
        """


class IZopeTree(INode):

    def __init__(root_object, id_attr='', children_attr='', request=None,
                 request_variable='', expanded_nodes=[], set_cookie=1):
        """
        'root_object' is the root object of the tree.

        'id_attr' and 'children_attr': see docstring for
        Node.__init__().

        'request' is the HTTP request.

        'request_variable' is the name of a request variable that
        contains the tree expansion information.
        """

    def encodeTreeExpansion(expanded_nodes):
        """
        Encode the tree expansion information in 'expanded_nodes'.
        """

    def decodeTreeExpansion(tree_expansion):
        """
        Decode the tree expansion information 'tree_expansion'.
        """

    def getFlatDicts():
        """
        Works like Node.getFlatNodes() except that it does not return
        a list of nodes but a list of dictionaries.
        """



"""Test suite."""
from Products.ZopeTree import Node, ZopeTree
from Products.ZopeTree.IZopeTree import INode, IZopeTree
from six import BytesIO
from Testing import ZopeTestCase
from ZPublisher.HTTPRequest import HTTPRequest
from ZPublisher.HTTPResponse import HTTPResponse
from ZTUtils.Tree import b2a
import unittest
import zlib


ZopeTestCase.installProduct('ZopeTree')


# simple object that can have an id and a set of children
class Item:
    def __init__(self, id, children=[]):
        self.id = id
        self.children = children


# function used to convert a set of nested tuples to items and
# children items.
def make_item_from_tuple(item_tuple, dict):
    children = []
    if len(item_tuple) > 1:
        for child in item_tuple[1]:
            children.append(make_item_from_tuple(child, dict))
    item = Item(item_tuple[0], children)
    dict[item_tuple[0]] = item
    return item


tree = ('a', [
    ('b', [('d',), ('e',)]), ('c', [('f', [('h',), ('i',)]), ('g')])])


expanded_nodes = ['a', 'c']


class NodeTest(ZopeTestCase.ZopeTestCase):

    def afterSetUp(self):
        # this mapping will provide shortcuts to each object
        self.items = {}
        self.root_obj = make_item_from_tuple(tree, self.items)
        self.root_node = Node(self.root_obj, 0, 'id', 'children',
                              expanded_nodes)

    def afterClear(self):
        pass

    def test_verify_interface(self):
        """Verify interface implementation"""
        from zope.interface.verify import verifyObject
        self.assertTrue(verifyObject(INode, self.root_node))

    def test_expand_collapse(self):
        """Test node expansion/collapsion"""
        root_node = self.root_node
        # first the node is expanded
        self.assertTrue(root_node.expanded)
        # now collapse it
        root_node.collapse()
        self.assertFalse(root_node.expanded)
        # make sure there are no children nodes returned!
        self.assertEqual(root_node.getChildrenNodes(), [])
        # expand it again
        root_node.expand()
        self.assertTrue(root_node.expanded)

    def test_children(self):
        """Test children"""
        root_node = self.root_node
        # testing hasChildren()
        self.assertTrue(root_node.hasChildren())

        # testing getChildrenNodes()
        children = [node.object for node in root_node.getChildrenNodes()]
        expected = [self.items['b'], self.items['c']]
        self.assertEqual(children, expected)

    def test_flat(self):
        """Test flat list generation"""
        root_node = self.root_node
        # testing getFlatNodes()
        children = [node.object for node in root_node.getFlatNodes()]
        # 'a' is not expected because the node on which getFlatNodes()
        # is called is not in the list
        expected = [self.items[i] for i in "bcfg"]
        self.assertEqual(children, expected)


class ZopeTreeTest(ZopeTestCase.ZopeTestCase):

    def afterSetUp(self):
        environ = {'SERVER_NAME': "", 'SERVER_PORT': '0'}
        response = HTTPResponse(stdout=BytesIO())
        request = HTTPRequest(BytesIO(), environ, response)
        self.varname = 'tree-expansion'
        # emulate a cookie
        tree_expansion = (":".join(expanded_nodes).encode("utf-8"))
        request.other[self.varname] = b2a(
            zlib.compress(tree_expansion)
            )
        self.request = request
        self.items = {}
        self.root_obj = make_item_from_tuple(tree, self.items)
        self.tree = ZopeTree(self.root_obj, 'id', 'children', request,
                             self.varname)

    def afterClear(self):
        pass

    def test_verify_interface(self):
        """Verify interface implementation"""
        from zope.interface.verify import verifyObject
        self.assertTrue(verifyObject(IZopeTree, self.tree))

    def test_encode_tree_expansion(self):
        """Test tree expansion encoding and decoding"""
        encoded = self.tree.encodeTreeExpansion(expanded_nodes)
        decoded = self.tree.decodeTreeExpansion(encoded)
        self.assertEqual(expanded_nodes, decoded)

    def test_pre_expanded(self):
        """Test pre-expanded nodes"""
        # 'a' is not expected because the node on which getFlatNodes()
        # is called is not in the list
        expected = [self.items[i] for i in "bcfg"]
        # test against to getFlatNodes()
        flat = [node.object for node in self.tree.getFlatNodes()]
        self.assertEqual(flat, expected)

    def test_flat_dicts(self):
        """Test flat dicts"""
        flatdicts = self.tree.getFlatDicts()
        self.assertEqual(len(flatdicts), len(self.tree.getFlatNodes()))
        bdict = flatdicts[0]
        self.assertEqual(bdict['id'], 'b')
        self.assertFalse(bdict['expanded'])
        self.assertEqual(bdict['depth'], 1)
        self.assertTrue(bdict['children'])
        self.assertTrue(bdict['object'] is self.items['b'])

    def test_cookie(self):
        """Test cookies"""
        # by default, the tree sets a cookie, so test for that
        request = self.request
        response = request.RESPONSE
        self.assertTrue(self.varname in response.cookies)

        # now make a tree that doesn't set a cookie
        treeexp = response.cookies[self.varname]['value']
        environ = {'SERVER_NAME': "", 'SERVER_PORT': '0'}
        response = HTTPResponse(stdout=BytesIO())
        request = HTTPRequest(BytesIO(), environ, response)
        request.other[self.varname] = treeexp
        self.tree = ZopeTree(self.root_obj, 'id', 'children', request,
                             self.varname, set_cookie=0)
        self.assertFalse(self.varname in response.cookies)


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(NodeTest))
    suite.addTest(unittest.makeSuite(ZopeTreeTest))
    return suite



from os import path
from setuptools import find_packages
from setuptools import setup

setup(name='Products.ZopeTree',
      version='2.0.1',
      url='https://github.com/jugmac00/Products.ZopeTree',
      project_urls={
          'Issue Tracker': 'https://github.com/jugmac00/Products.ZopeTree/issues',
          'Sources': 'https://github.com/jugmac00/Products.ZopeTree',
      },      
      description="ZopeTree is a light-weight tree implementation.",
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      maintainer='Juergen Gmach',
      maintainer_email="juergen.gmach@googlemail.com",
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Zope :: 4',
        'Framework :: Zope :: 5',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Mozilla Public License 1.1 (MPL 1.1)',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
      ],
      packages=find_packages('src'),
      namespace_packages=['Products'],
      package_dir={'': 'src'},
      python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*,!=3.4.*,!=3.5.*',
      install_requires=[
        'Zope',
        'setuptools',
        'six',
        'AccessControl',
        'zope.interface',
        'zope.testing',
      ],
      include_package_data=True,
      zip_safe=False,
      options={"bdist_wheel": {"universal": "1"}},
      )



#
# ZopeTree
#
# Simple tree implementation for Zope
#
# Copyright (c) 2001-2003 by Philipp "philiKON" von Weitershausen
#
# This software is distributed under the terms of the Mozilla Public
# License (MPL)
#

"""
$Id: ZopeTree.py,v 1.8 2003/05/30 15:13:02 philipp Exp $
"""

import zlib
import zope.interface
from ZTUtils.Tree import b2a, a2b

from Products.ZopeTree.IZopeTree import IZopeTree, INode


@zope.interface.implementer(INode)
class Node:

    __allow_access_to_unprotected_subobjects__ = 1

    def _get_attr(self, attr):
        """
        Get the attribute 'attr' from the wrapped object. If the
        result is callable, call it.
        """
        attr = getattr(self.object, attr)
        if callable(attr):
            return attr()
        else:
            return attr

    def _create_children_nodes(self):
        """
        Create children nodes
        """
        try:
            children = self._get_attr(self._children_attr)
        # yes, this is bad, but _get_attr can throw almost all exceptions
        except:  # noqa
            self._children_nodes = []
            return
        nodes = []
        for obj in children:
            node = Node(obj, self.depth+1, self._id_attr,
                        self._children_attr, self._expanded_nodes)
            nodes.append(node)
        self._children_nodes = nodes

    def __init__(self, object, depth, id_attr, children_attr,
                 expanded_nodes=[]):
        # attributes required by the interface
        self.object = object
        self.depth = depth
        self.expanded = 0

        self._id_attr = id_attr
        self._children_attr = children_attr
        self._expanded_nodes = expanded_nodes

        if self.getId() in expanded_nodes:
            self.expand()

    def expand(self, recursive=0):
        self.expanded = 1
        if recursive:
            for node in self.getChildrenNodes():
                node.expand(1)

    def collapse(self):
        self.expanded = 0

    def getId(self):
        return self._get_attr(self._id_attr)

    def hasChildren(self):
        try:
            return len(self._get_attr(self._children_attr)) > 0
        except AttributeError:
            return 0

    def getChildrenNodes(self):
        if not self.expanded:
            return []
        # children nodes are not created until they are explicitly requested
        # through this function
        if not hasattr(self, '_children_nodes'):
            self._create_children_nodes()
        return self._children_nodes[:]

    def getFlatNodes(self):
        nodes = []
        for node in self.getChildrenNodes():
            nodes.append(node)
            nodes += node.getFlatNodes()
        return nodes


def safe_decompress(input, max_size=10240):
    # this sillyness can go away in python 2.2
    decomp = zlib.decompressobj()
    output = ''
    while input:
        fragment_size = max(1, (max_size-len(output))/1000)
        fragment, input = input[:fragment_size], input[fragment_size:]
        output += decomp.decompress(fragment)
        if len(output) > max_size:
            raise ValueError('Compressed input too large')
    return output + decomp.flush()


@zope.interface.implementer(IZopeTree)
class ZopeTree(Node):

    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, root_object, id_attr='getId',
                 children_attr='objectValues', request=None,
                 request_variable='tree-expansion', expanded_nodes=[],
                 set_cookie=1):
        tree_expansion = request.get(request_variable, "")
        if tree_expansion:
            if set_cookie:
                # set a cookie right away
                request.RESPONSE.setCookie(request_variable, tree_expansion)
            expanded_nodes = self.decodeTreeExpansion(tree_expansion)

        Node.__init__(self, root_object, 0, id_attr, children_attr,
                      expanded_nodes)
        self.expand()

    def encodeTreeExpansion(self, expanded_nodes):
        tree_expansion = ":".join(expanded_nodes)
        if not isinstance(tree_expansion, bytes):
            tree_expansion = tree_expansion.encode("utf-8")
        tree_expansion = zlib.compress(tree_expansion)
        return b2a(tree_expansion)

    def decodeTreeExpansion(self, tree_expansion):
        tree_expansion = a2b(tree_expansion)
        tree_expansion = zlib.decompress(tree_expansion)
        if not isinstance(tree_expansion, str):
            tree_expansion = tree_expansion.decode('utf-8')
        return tree_expansion.split(":")

    def getFlatDicts(self):
        flatdicts = []
        self.maxdepth = 0
        for node in self.getFlatNodes():
            id = node.getId()
            expanded_nodes = self._expanded_nodes[:]
            if id in self._expanded_nodes:
                # if the node is already expanded, the next step is
                # collapsing it
                expanded_nodes.remove(id)
            else:
                # if it isn't, the next step is expanding it.
                expanded_nodes += [id]
            flatdicts.append({
                'id':             id,
                'expanded':       node.expanded,
                'depth':          node.depth,
                'children':       node.hasChildren(),
                'tree-expansion': self.encodeTreeExpansion(expanded_nodes),
                'object':         node.object
                })
            if node.depth > self.maxdepth:
                self.maxdepth = node.depth
        return flatdicts
